#------------------------------------------------------------------------------------------------
# KLayout forum No.1642
#
# Sub: Python Scripting for reading an Excel file and creating a polar array on Klayout Macro
# Posted by: dsenuka
#        on: September 16
#
# Author       : Kazzz-S
# Last modified: 2020-09-28
#------------------------------------------------------------------------------------------------
# Usage:
#        (1) Store this script in ${HOME}/.klayout/pymacros/
#        (2) Start KLayout and open a New Layout
#        (3) Set the top cell name 'metalens'; database unit=MyDBU [um]
#        (4) Start the Macro Development IDE tool
#        (5) Open this script and run
#------------------------------------------------------------------------------------------------
import os
import platform
import math
import pandas as pd
import pya

#-------------------------------------------------------------------------------------
# Global control parameters; modify and set these parameters as you like
#-------------------------------------------------------------------------------------
Debug          = True   # set False not to print debug messages
NumVertices    = None   # set an integer like 32 or 16 to fix the number of vertices
                        # of a polygon that approximates a circle
Convert2Static = False  # set True to convert PCells to static cells
MyDBU          = 0.001  # set an appropriate database unit in [um]
NumDigits      = 8      # set an appropriate number of digits below the decimal point
MyRoot         = os.environ["HOME"] + "/KLayout/Forum1642/"
MyCSV          = MyRoot + "RealKL1642.csv"
OutGDS2        = MyRoot + "RealKL1642.gds"

#-------------------------------------------------------------------------------------
## Compute the number of vertices of a regular-N polygon approximating a circle
#
# @param[in] radius     radius in [um]
# @param[in] tol        tolerance in [um] (default=0.001[um])
# @param[in] integral4  the number should be an integral multiple of four (default=True)
#
# @return the number of vertices
#-------------------------------------------------------------------------------------
def ComputeNumberVerticies( radius, tol=0.001, integral4=True ):
  global NumVertices

  if not NumVertices == None:
    return NumVertices

  x = 1.0 - tol / radius
  y = math.acos(x)
  z = int( 2*math.pi / y )
  if not integral4:
    numver = z
  else:
    res = z % 4
    if   res == 0:
      numver = z
    elif res == 1:
      numver = z + 3
    elif res == 2:
      numver = z + 2
    elif res == 3:
      numver = z + 1
    else:
      numver = 0 # dummy
  return numver

#-------------------------------------------------------------------------------------
# KLayout's major objects
#-------------------------------------------------------------------------------------
MainWindow  = pya.MainWindow.instance()
CurrentView = MainWindow.current_view()
Layout      = CurrentView.active_cellview().layout()
DesignFname = CurrentView.active_cellview().filename()
TopCellPtr  = Layout.top_cell()
TopCellName = TopCellPtr.name   # Python getter
CurrentDBU  = Layout.dbu        # Python getter; database unit
if not CurrentDBU == MyDBU:
  raise Exception( "! MyDBU=%f but CurrentDBU=%f" % (MyDBU, CurrentDBU) )

#---------------------------------------------------------------------------------------
## Construct the circle dictionary by reading the (modified) CSV file with four columns
#
#    #------------------------------------------------------------------------------------------------
#    # KLayout forum No.1642
#    #
#    # Sub: Python Scripting for reading an Excel file and creating a polar array on Klayout Macro
#    # Posted by: dsenuka
#    #        on: September 16
#    #
#    # Source       : all_data_table_wavelength_4um_lens_radius_2000um_f#_2_with_min_and_max_angles.xlsx
#    # Remade by    : Kazzz-S
#    # Last modified: 2020-09-27
#    #------------------------------------------------------------------------------------------------
#    pillar number,x coordinate [um],y coordinate [um],orbit radius [um],circle diameter [um],rotation angle [rad],number of pillars to built (to make a quarter),From angle [rad],Up to angle [rad],Up to angle [rad]
#    1,0.75000000 ,0.75000000 ,1.06066017 ,1.09998989 ,1.57079633 ,1 ,0.78539816 ,7.854E-01,0.78539816
#    2,2.25000000 ,0.93198052 ,2.43538245 ,1.09990900 ,0.78539816 ,2 ,0.39269908 ,1.178E+00,1.17809725
#    :
#    :
#    1333,"1,998.75000000 ",0.75003089 ,"1,998.75014072 ",0.62390673 ,0.00075050 ,"2,093 ",0.00037525 ,1.570E+00,1.57042108
#    1334,"2,000.25000000 ",0.75023532 ,"2,000.25014070 ",1.06437124 ,0.00075014 ,"2,094 ",0.00037507 ,1.570E+00,1.57042126
#    1335,"2,001.75000000 ",0.75008151 ,"2,001.75014053 ",0.97396148 ,0.00074943 ,"2,096 ",0.00037471 ,1.570E+00,1.57042161
#
#---------------------------------------------------------------------------------------
def ConstructCircleDictionary():
  global CircleDic    # the circle dictionary

  CircleDic   = dict()
  df          = pd.read_csv( MyCSV, comment='#' )
  numLeafCell = len(df)
  for idx in range(0, numLeafCell):
    cirID    = int  ( df.iloc[idx,0] )
    startX   = float( df.iloc[idx,1] )
    startY   = float( df.iloc[idx,2] )
    orbitRad = float( df.iloc[idx,3] )
    cirDia   = float( df.iloc[idx,4] )
    rotAngle = float( df.iloc[idx,5] )
    numInst  = int  ( df.iloc[idx,6] )
    CircleDic[cirID] = (startX, startY, orbitRad, cirDia, rotAngle, numInst)

  if Debug:
    print( "### Contents of CircleDic ###" )
    for cirID in sorted(CircleDic.keys()):
      print( cirID, CircleDic[cirID] )
    print( "" )

#-------------------------------------------------------------------------------
## Create different circles as PCells and populate the cell dictionary
#-------------------------------------------------------------------------------
def CreateLeafPCells():
  global CellDic    # the cell dictionary

  CellDic = dict()

  #-----------------------------------------------------------------------
  # [1] Create each PCell as "CIRCLE" (not "ELLIPSE")
  #-----------------------------------------------------------------------
  for cirID in sorted(CircleDic.keys()):
    #---------------------------------------------------------------------
    # (A) Populate a dictionary to pass PCell parameters
    #---------------------------------------------------------------------
    (startX, startY, orbitRad, cirDia, rotAngle, numInst) = CircleDic[cirID]

    cellName = "C%04d" % cirID # max. is 1335
    numVer   = ComputeNumberVerticies( round( cirDia / 2.0, NumDigits ), tol=MyDBU )

    paraL = "layer"         # Layer and Datatype
    paraR = "radius"        # radius shown in the dialog for further CAD operation
    paraA = "actual_radius" # hidden parameter that actually changes the radius
    paraN = "npoints"       # number of vertices of a polygon approximating the circle

    params = dict()
    params[paraL] = pya.LayerInfo( 1, 0 ) # use a single layer for simplicity
    params[paraR] = round( cirDia / 2.0, NumDigits )
    params[paraA] = round( cirDia / 2.0, NumDigits )
    params[paraN] = numVer

    #---------------------------------------------------------------------
    # (B) Create a PCell and convert it to a static cell on demand
    #---------------------------------------------------------------------
    sbcell1 = Layout.create_cell( "CIRCLE", "Basic", params ).cell_index()
    if Convert2Static:
      sbcell2 = Layout.convert_cell_to_static( sbcell1 )
      Layout.rename_cell( sbcell2, cellName )
      Layout.delete_cell( sbcell1 )
      CellDic[cirID] = sbcell2
      if Debug:
        print( "Leaf static cell <%s> has Cell-ID <%d>" % (cellName, sbcell2) )
    else:
      Layout.rename_cell( sbcell1, cellName )
      CellDic[cirID] = sbcell1
      if Debug:
        print( "Leaf PCell <%s> has Cell-ID <%d>" % (cellName, sbcell1) )

  #-----------------------------------------------------------------------
  # [2] Take care of the newly added layers
  #-----------------------------------------------------------------------
  CurrentView.add_missing_layers()  # add missing layers
  CurrentView.zoom_fit()            # zoom to fit the entire hierarchy

#-------------------------------------------------------------------------------
## Create the circle array in a quarter (quadrant-I)
#-------------------------------------------------------------------------------
def CreateCircleArray():
  global TotalInstances   # total number of instances inserted

  TotalInstances = 0
  for cirID in sorted(CircleDic.keys()):

    #-----------------------------------------------------
    # [1] Look up the dictionaries
    #-----------------------------------------------------
    (startX, startY, orbitRad, cirDia, rotAngle, numInst) = CircleDic[cirID]
    startAngle = math.atan2( startY, startX )
    leafCell   = CellDic[cirID]

    #-----------------------------------------------------
    # [2] Place the leaf cell one by one
    #-----------------------------------------------------
    for idx in range(0, numInst):
      #-----------------------------------------------------
      # (A) Set the location on the orbit
      #-----------------------------------------------------
      if idx == 0:
        x = round( startX, NumDigits )
        y = round( startY, NumDigits )
      else:
        theta = startAngle + idx * rotAngle
        x     = round( orbitRad * math.cos(theta), NumDigits )
        y     = round( orbitRad * math.sin(theta), NumDigits )

      #-----------------------------------------------------
      # (B) SREF the PCell
      #-----------------------------------------------------
      dispVec = pya.DTrans( x/MyDBU, y/MyDBU )  # displace cell by (x, y) [um]
      cellIns = pya.CellInstArray()             # default cell instance
      cellIns.cell_index = leafCell             # 'set' the cell to refer (what to refer)
      cellIns.trans = dispVec                   # 'set' the displacement vector (how to refer)
      TopCellPtr.insert( cellIns )              # insert the instance to [TopCellName]
      TotalInstances += 1

#-------------------------------------------------------------------------------
## Generate the GDS2 file
#-------------------------------------------------------------------------------
def GenerateGDS2():
  sop = pya.SaveLayoutOptions()   # default save option
  Layout.write( OutGDS2, sop )    # *.gds makes KLayout infer the output format
  print( "" )
  print( "### Generated", OutGDS2 )

#===============================================================================
# Main()
#===============================================================================

#-------------------------------------------------------------
# [0] Get machine environment and start the timer
#-------------------------------------------------------------
(System, Node, Release, Version, Machine, Processor) = platform.uname()
myTimer = pya.Timer()
myTimer.start()

#-------------------------------------------------------------
# [1] Construct the cell hierarchy
#-------------------------------------------------------------
ConstructCircleDictionary()
CreateLeafPCells()
CreateCircleArray()

#-------------------------------------------------------------
# [2] Change the hierarchy level
#-------------------------------------------------------------
CurrentView.max_hier_levels = 2
CurrentView.zoom_fit()

#-------------------------------------------------------------
# [3] Generate the GDS2 file
#-------------------------------------------------------------
GenerateGDS2()

#-------------------------------------------------------------
# [4] Print timing info
#-------------------------------------------------------------
myTimer.stop()
usingPCell = { True:"No", False:"Yes" } # key=Convert2Static
print( "" )
print( "### <%s> processed <%s> ###" % ( os.path.basename(__file__), os.path.basename(MyCSV) ) )
print( "    KLayout version: %s" % pya.Application().instance().version() )
print( "    Machine info: %s, %s, %s" % (System, Node, Machine) )
print( "    Using PCell?: %s" % usingPCell[ Convert2Static ] )
print( "    Total number of instances inserted: %s" % "{:,}".format(TotalInstances) )
print( "    Timing info: Sys=%.3f[sec]  User=%.3f[sec]  Wall=%.3f[sec]" % ( myTimer.sys(), myTimer.user(), myTimer.wall() ) )
print( "" )

#-------------------------------------------------------------
# [5] Manual operations can follow...
#-------------------------------------------------------------
# You can
#     (A) change the layer properties
#     (B) convert a PCell to a static cell, if not yet done
#     (C) flat the cell hierarchy
#     (D) change the layers
#     (E) and many more...
#
# Enjoy PYA scripting!
#-------------------------------------------------------------

#---------------
# End of file
#---------------
